import pytest
from SCNIC.module_maker import make_modules, collapse_modules, write_modules_to_dir
import os
import glob
from biom.table import Table
import numpy as np
from scipy.spatial.distance import squareform


@pytest.fixture()
def biom_table1():
    arr = np.array([[250,   0, 100, 446,   75],
                    [  0,   0,   1,   1,    2],
                    [  2,   2,   2,   2,    2],
                    [100, 100, 500,   1, 1000],
                    [500,   5,   0,  50,  100]])
    obs_ids = ["otu_%s" % i for i in xrange(5)]
    samp_ids = ["samp_%s" % i for i in xrange(5)]
    return Table(arr, obs_ids, samp_ids)


@pytest.fixture()
def dist1():
    dist = np.array([[0.0, .05, 0.2, 1.0, 1.0],
                     [.05, 0.0, 0.2, 1.0, 1.0],
                     [0.2, 0.2, 0.0, 1.0, 1.0],
                     [1.0, 1.0, 1.0, 0.0, 0.7],
                     [1.0, 1.0, 1.0, 0.7, 0.0]])
    return squareform(dist)


@pytest.fixture()
def obs_ids1():
    obs_ids = ["otu_%s" % i for i in xrange(5)]
    return obs_ids


@pytest.fixture()
def modules1(dist1, obs_ids1):
    return make_modules(dist1, min_dist=.21, obs_ids=obs_ids1)


def test_make_modules(modules1):
    assert len(modules1) == 1
    assert np.sum([len(i) for i in modules1]) == 3


def test_collapse_modules(biom_table1, modules1):
    coll_table = collapse_modules(biom_table1, modules1)
    assert coll_table.shape[0] == biom_table1.shape[0]-2
    assert coll_table.shape[1] == biom_table1.shape[1]
    assert coll_table.sum() == biom_table1.sum()
    assert np.array_equal(coll_table.sum(axis="sample"), biom_table1.sum(axis="sample"))
    # set since order doesn't matter
    assert frozenset(coll_table.sum(axis="observation")) == frozenset(np.array([885, 1701, 655]))


def test_write_modules_to_dir(biom_table1, modules1, tmpdir):
    tempdir = tmpdir.mkdir("mods")
    os.chdir(str(tempdir))
    write_modules_to_dir(biom_table1, modules1)
    os.chdir("modules")
    fnames = glob.glob(str(tempdir)+"/modules/*.biom")
    assert len(fnames) == len(modules1)



"""Make modules of observations based on cooccurence networks and collapse table"""

from __future__ import division

import numpy as np
import pandas as pd
from biom import Table
import os
from scipy.cluster.hierarchy import complete
from scipy.spatial.distance import squareform
from skbio.tree import TreeNode
from itertools import combinations


def make_modules(dist, min_dist, obs_ids):
    # create linkage matrix using complete linkage
    Z = complete(dist)
    # make tree from linkage matrix with names from dist
    tree = TreeNode.from_linkage_matrix(Z, obs_ids)
    # get all tips so in the end we can check if we are done
    all_tips = len([i for i in tree.postorder() if i.is_tip()])
    cliques = set()
    seen = set()
    dist = pd.DataFrame(squareform(dist), index=obs_ids, columns=obs_ids)
    for node in tree.levelorder():
        if node.is_tip():
            seen.add(node.name)
        else:
            tip_names = frozenset((i.name for i in node.postorder() if i.is_tip()))
            if tip_names.issubset(seen):
                continue
            dists = (dist.loc[tip1, tip2] > min_dist for tip1, tip2 in combinations(tip_names, 2))
            if any(dists):
                continue
            else:
                cliques.add(tip_names)
                seen.update(tip_names)
        if len(seen) == all_tips:
            return cliques
    raise ValueError("Well, how did I get here?")


def collapse_modules(table, modules, prefix="module"):
    """collapse created modules in a biom table, members of multiple modules will be added to the smallest module"""
    table = table.copy()
    module_array = np.zeros((len(modules), table.shape[1]))

    seen = set()
    for i, module in enumerate(modules):
        seen = seen | module
        # sum everything in the module
        module_array[i] = np.sum([table.data(feature, axis="observation") for feature in module], axis=0)

    table.filter(seen, axis='observation', invert=True)

    # make new table
    new_table_matrix = np.concatenate((table.matrix_data.toarray(), module_array))
    new_table_obs = list(table.ids(axis='observation')) + ['_'.join([prefix, str(i)]) for i in xrange(len(modules))]
    return Table(new_table_matrix, new_table_obs, table.ids())


def write_modules_to_dir(table, modules):
    # for each module merge values and print modules to file
    os.makedirs("modules")
    # reverse modules so observations will be added to smallest modules
    for i, module in enumerate(modules):
        # make biom tables for each module and write to file
        module_table = table.filter(module, axis='observation', inplace=False)
        module_table.to_json("modulemaker.py", open("modules/%s.biom" % i, 'w'))


def write_modules_to_file(modules, prefix="module"):
    # write all modules to file
    with open("modules.txt", 'w') as f:
        for i, module in enumerate(modules):
            f.write('_'.join([prefix, str(i)]) + '\t' + '\t'.join([str(j) for j in module]) + '\n')



"""
Workflow script for finding correlations between pairs of biom tables, making networks, finding modules and collapsing
modules.
"""
import os
from biom import load_table
from scipy.stats import spearmanr, pearsonr
import networkx as nx
import numpy as np
from SCNIC import general
from SCNIC.correlation_analysis import between_correls_from_tables
import shutil

__author__ = 'shafferm'

# TODO: output heat map with clusters

def between_correls(args):
    """TABLES MUST SORT SO THAT SAMPLES ARE IN THE SAME ORDER """
    logger = general.Logger("SCNIC_log.txt")
    logger["SCNIC analysis type"] = "between"

    # correlation and p-value adjustment methods
    correl_methods = {'spearman': spearmanr, 'pearson': pearsonr}
    p_methods = {'bh': general.bh_adjust, 'bon': general.bonferroni_adjust}
    correl_method = correl_methods[args.correl_method]
    if args.p_adjust is not None:
        p_adjust = p_methods[args.p_adjust]
    else:
        p_adjust = None

    # load tables
    table1 = load_table(args.table1)
    table2 = load_table(args.table2)
    logger["input table 1"] = args.table1
    logger["input table 1"] = args.table2

    table1 = table1.sort()
    table2 = table2.sort()

    # make new output directory and change to it
    if args.force and args.output is not None:
        shutil.rmtree(args.output, ignore_errors=True)
    if args.output is not None:
        os.makedirs(args.output)
        os.chdir(args.output)
        logger["output directory"] = args.output

    # filter tables
    if args.sparcc_filter is True:
        table1 = general.sparcc_paper_filter(table1)
        table2 = general.sparcc_paper_filter(table2)
        print "Table 1 filtered: " + str(table1.shape[0]) + " observations"
        print "Table 2 filtered: " + str(table2.shape[0]) + " observations"
        logger["sparcc paper filter"] = True
        logger["number of observations present in table 1 after filter"] = table1.shape[0]
        logger["number of observations present in table 2 after filter"] = table2.shape[0]
    if args.min_sample is not None:
        table1 = general.filter_table(table1, args.min_sample)
        table2 = general.filter_table(table2, args.min_sample)

    if not np.array_equal(table1.ids(), table2.ids()):
        raise ValueError("Tables have different sets of samples present")

    metadata = general.get_metadata_from_table(table1)
    metadata.update(general.get_metadata_from_table(table2))

    # make correlations
    logger["correlation metric"] = args.correl_method
    logger["p adjustment method"] = args.p_adjust
    correls = between_correls_from_tables(table1, table2, correl_method, nprocs=args.procs)
    correls.sort_values(correls.columns[-1], inplace=True)
    correls['p_adj'] = p_adjust(correls['p'])
    correls.to_csv(open('correls.txt', 'w'), sep='\t', index=False)

    # make network
    net = general.correls_to_net(correls, metadata=metadata, min_p=args.min_p, min_r=args.min_r)
    logger["number of nodes"] = net.number_of_nodes()
    logger["number of edges"] = net.number_of_edges()
    nx.write_gml(net, 'crossnet.gml')

    logger.output_log()
    print '\a'



import pytest
import os
from SCNIC.general import simulate_correls
from SCNIC.between_correls import between_correls


@pytest.fixture()
def args():
    class Arguments(object):
        def __init__(self):
            self.table1 = "table1.biom"
            self.table2 = "table2.biom"
            self.output = "out_dir"
            self.correl_method = "spearman"
            self.p_adjust = "bh"
            self.min_sample = None
            self.min_p = None
            self.min_r = None
            self.sparcc_filter = False
            self.force = False
            self.procs = 1

    return Arguments()


def test_between_correls(args, tmpdir):
    table1 = simulate_correls()
    table2 = simulate_correls()
    loc = tmpdir.mkdir("with_correls_test")
    table1.to_json("madebyme", open(str(loc)+"/table1.biom", 'w'))
    table2.to_json("madebyme", open(str(loc) + "/table2.biom", 'w'))
    os.chdir(str(loc))
    between_correls(args)
    files = os.listdir(str(loc)+'/out_dir')
    assert "correls.txt" in files
    assert "crossnet.gml" in files



#!/usr/local/bin/python2

import argparse
from SCNIC.within_correls import within_correls
from SCNIC.between_correls import between_correls

__author__ = 'shafferm'

"""Entry to both module_maker and between_correls, only holds main and args are passed to the corresponding program
"""


def main():
    """Things"""
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers()
    within_corr = subparsers.add_parser("within", help="Find pairwise correlations within a table and make modules",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    between_corr = subparsers.add_parser("between", help="Find correlations between two tables",
                                         formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    within_corr.add_argument("-i", "--input", help="location of input biom file", required=True)
    within_corr.add_argument("-o", "--output", help="output file location")
    within_corr.add_argument("-m", "--correl_method", help="correlation method", default="sparcc")
    within_corr.add_argument("-a", "--p_adjust", help="p-value adjustment", default="bh")
    within_corr.add_argument("-s", "--min_sample", help="minimum number of samples present in", type=int)
    within_corr.add_argument("--prefix", help="prefix for module names in collapsed file", default="module_")
    within_corr.add_argument("--min_p", help="minimum p-value to determine edges", type=float)
    within_corr.add_argument("--min_r", help="minimum correlation value to determine edges", type=float)
    within_corr.add_argument("--procs", help="number of processors to use", default=1, type=int)
    within_corr.add_argument("-b", "--bootstraps", help="number of bootstraps", default=100, type=int)
    within_corr.add_argument("-f", "--force", help="force overwrite output folder if it already exists", default=False,
                             action="store_true")
    within_corr.add_argument("--sparcc_filter", help="filter as described in SparCC paper", default=False,
                             action="store_true")
    within_corr.add_argument("--verbose", help="give verbose messages to STDOUT", default=False, action="store_true")
    within_corr.set_defaults(func=within_correls)

    between_corr.add_argument("-1", "--table1", help="table to be correlated", required=True)
    between_corr.add_argument("-2", "--table2", help="second table to be correlated", required=True)
    between_corr.add_argument("-o", "--output", help="output file location")
    between_corr.add_argument("-m", "--correl_method", help="correlation method", default="spearman")
    between_corr.add_argument("-a", "--p_adjust", help="p-value adjustment", default="bh")
    between_corr.add_argument("-s", "--min_sample", help="minimum number of samples present in", type=int)
    between_corr.add_argument("--min_p", help="minimum p-value to determine edges", type=float)
    between_corr.add_argument("--min_r", help="minimum R to determine edges", type=float)
    between_corr.add_argument("--sparcc_filter", help="filter using parameters from SparCC publication", default=False,
                              action="store_true")
    between_corr.add_argument("--procs", help="number of processors to use", default=1, type=int)
    between_corr.add_argument("-f", "--force", help="force overwrite output folder if it already exists", default=False,
                              action="store_true")
    between_corr.set_defaults(func=between_correls)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



from __future__ import division

from scipy.stats import linregress
import numpy as np
import networkx as nx
from biom.table import Table
import pandas as pd
from datetime import datetime
from collections import OrderedDict
from numpy.random import multivariate_normal


__author__ = 'shafferm'


"""functions used widely"""


class Logger(OrderedDict):
    """"""
    # TODO: break up into sections for correls making, network making and module making
    def __init__(self, output):
        super(Logger, self).__init__()
        self.output_file = output
        self['start time'] = datetime.now()

    def output_log(self):
        with open(self.output_file, 'w') as f:
            self['finish time'] = datetime.now()
            self['elapsed time'] = self['finish time'] - self['start time']
            for key, value in self.iteritems():
                f.write(key + ': ' + str(value) + '\n')


def sparcc_paper_filter(table):
    """if a observation averages more than 2 reads per sample then keep,
    if a sample has more than 500 reads then keep"""
    table = table.copy()
    table.filter(table.ids(axis='sample')[table.sum(axis='sample') > 500], axis='sample')
    table.filter(table.ids(axis='observation')[table.sum(axis='observation') / table.shape[1] >= 2], axis="observation")
    return table


def df_to_biom(df):
    return Table(np.transpose(df.as_matrix()), list(df.columns), list(df.index))


def biom_to_df(biom):
    return pd.DataFrame(np.transpose(biom.matrix_data.todense()), index=biom.ids(),
                        columns=biom.ids(axis="observation"))


def get_metadata_from_table(table):
    metadata = dict()
    for _, otu_i, metadata_i in table.iter(axis="observation"):
        if metadata_i is not None:
            metadata[otu_i] = metadata_i
    return metadata


def bh_adjust(pvalues):
    """
    benjamini-hochberg p-value adjustment stolen from
    http://stackoverflow.com/questions/7450957/how-to-implement-rs-p-adjust-in-python

    Parameters
    ----------
    pvalues: an iterable of p-values

    Returns
    -------
    new_pvalues: a numpy array of BH adjusted p-values
    """
    pvalues = np.array(pvalues)
    n = pvalues.shape[0]
    new_pvalues = np.empty(n)
    values = [(pvalue, i) for i, pvalue in enumerate(pvalues)]
    values.sort()
    values.reverse()
    new_values = []
    for i, vals in enumerate(values):
        rank = n - i
        pvalue, index = vals
        new_values.append((n/rank) * pvalue)
    for i in xrange(0, int(n)-1):
        if new_values[i] < new_values[i+1]:
            new_values[i+1] = new_values[i]
    for i, vals in enumerate(values):
        pvalue, index = vals
        new_pvalues[index] = new_values[i]
    return new_pvalues


def bonferroni_adjust(pvalues):
    pvalues = np.array(pvalues)
    n = float(pvalues.shape[0])
    new_pvalues = n * pvalues
    return new_pvalues


def correls_to_net(correls, min_p=None, min_r=None, conet=False, metadata=None):
    """correls is a pandas dataframe which has columns feature1, feature2, r and optionally p and p_adj and optionally
    any others"""
    if metadata is None:
        metadata = []

    if min_p is None and min_r is None:
        min_p = .05

    if conet:
        correls = correls[correls.r > 0]

    if min_p is not None:
        # filter to only include significant correlations
        if 'p_adj' in correls.columns:
            correls = correls[correls.p_adj < min_p]
        elif 'p' in correls.columns:
            correls = correls[correls.p < min_p]
        else:
            raise ValueError("No p or p_adj in correls")

    if min_r is not None:
        if conet:
            correls = correls[correls.r > min_r]
        else:
            correls = correls[np.abs(correls.r) > min_r]

    graph = nx.Graph()
    for _, correl in correls.iterrows():
        graph.add_node(correl.feature1)
        if correl.feature1 in metadata:
            for key in metadata[correl.feature1]:
                graph_key = str(key).replace('_', '')
                if metadata[correl.feature1][key] is None:
                    continue
                if hasattr(metadata[correl.feature1][key], '__iter__'):
                    graph.node[correl.feature1][graph_key] = ';'.join(metadata[correl.feature1][key])
                else:
                    graph.node[correl.feature1][graph_key] = metadata[correl.feature1][key]

        graph.add_node(correl.feature2)
        if correl.feature2 in metadata:
            for key in metadata[correl.feature2]:
                graph_key = str(key).replace('_', '')
                if metadata[correl.feature2][key] is None:
                    continue
                if hasattr(metadata[correl.feature2][key], '__iter__'):
                    graph.node[correl.feature2][graph_key] = ';'.join(metadata[correl.feature2][key])
                else:
                    graph.node[correl.feature2][graph_key] = metadata[correl.feature2][key]
        graph.add_edge(correl.feature1, correl.feature2)
        for i in correl.index[2:]:
            graph_key = i.replace('_', '')
            graph.edge[correl.feature1][correl.feature2][graph_key] = correl[i]
    return graph


def filter_table(table, min_samples=None, to_file=False):
    """filter relative abundance table, by default throw away things greater than 1/3 zeros"""
    table = table.copy()
    # first sample filter
    if min_samples is not None:
        to_keep = [i for i in table.ids(axis='observation')
                   if sum(table.data(i, axis='observation') != 0) >= min_samples]
    else:
        to_keep = [i for i in table.ids(axis='observation')
                   if sum(table.data(i, axis='observation') != 0) >= table.shape[1]/3]
    table.filter(to_keep, axis='observation')

    if to_file:
        table.to_json('filter_table', open("filtered_tab.biom", 'w'))
        # open("filtered_rel_abund.txt", 'w').write(table.to_tsv())

    return table


def simulate_correls(corr_stren=(.99, .99), std=(1, 1, 1, 2, 2), means=(100, 100, 100, 100, 100), size=30,
                     noncors=10, noncors_mean=100, noncors_std=100, subsample=None, log=False):
    """
    Generates a correlation matrix with diagonal of stds based on input parameters and fills rest of matrix with
    uncorrelated values all with same  mean and standard deviations. Output should have a triangle of correlated
    observations and a pair all other observations should be uncorrelated. Correlation to covariance calculated by
    cor(X,Y)=cov(X,Y)/sd(X)sd(Y).

    Parameters
    ----------
    corr_stren: tuple of length 2, correlations in triangle and in pair
    std: tuple of length 5, standard deviations of each observation
    means: tuple of length 5, mean of each observation
    size: number of samples to generate from the multivariate normal distribution
    noncors: number of uncorrelated values
    noncors_mean: mean of uncorrelated values
    noncors_std: standard deviation of uncorrelated values
    subsample: Rarefy data to give threshold
    log: logonentiate mean values, if you are trying to detect log correlation

    Returns
    -------
    table: a biom table with (size) samples and (5+noncors) observations
    """
    cor = [[std[0], corr_stren[0], corr_stren[0], 0., 0.],  # define the correlation matrix for the triangle and pair
           [corr_stren[0], std[1], corr_stren[0], 0., 0.],
           [corr_stren[0], corr_stren[0], std[2], 0., 0.],
           [0., 0., 0., std[3], corr_stren[1]],
           [0., 0., 0., corr_stren[1], std[4]]]
    cor = np.array(cor)
    cov = np.zeros(np.array(cor.shape) + noncors)  # generate empty covariance matrix to be filled
    for i in xrange(cor.shape[0]):  # fill in all but diagonal of covariance matrix, first 5
        for j in xrange(i + 1, cor.shape[0]):
            curr_cov = cor[i, j] * cor[i, i] * cor[j, j]
            cov[i, j] = curr_cov
            cov[j, i] = curr_cov
    for i in xrange(cor.shape[0]):  # fill diagonal of covariance matrix, first 5
        cov[i, i] = np.square(cor[i, i])
    means = list(means)
    for i in xrange(cor.shape[0], cov.shape[0]):  # fill diagonal of covariance, 6 to end and populate mean list
        cov[i, i] = noncors_std
        means.append(noncors_mean)

    if log:
        # if log then log the array
        means = np.log(np.array(means))

    # fill the count table
    counts = multivariate_normal(means, cov, size).T

    if log:
        counts = np.log(counts)

    counts = np.round(counts)

    observ_ids = ["Observ_" + str(i) for i in xrange(cov.shape[0])]
    sample_ids = ["Sample_" + str(i) for i in xrange(size)]
    table = Table(counts, observ_ids, sample_ids)

    if subsample is not None:
        table = table.subsample(subsample)

    return table



import pytest
import os
from SCNIC.general import simulate_correls
from SCNIC.within_correls import within_correls


@pytest.fixture()
def args1():
    class Arguments(object):
        def __init__(self):
            self.input = "table1.biom"
            self.output = "out_dir"
            self.correl_method = "spearman"
            self.p_adjust = "bh"
            self.outlier_removal = False
            self.verbose = False
            self.force = False
            self.min_sample = 2
            self.sparcc_filter = False
            self.procs = 1
            self.min_p = None
            self.min_r = .35

    return Arguments()


@pytest.fixture()
def args2():
    class Arguments(object):
        def __init__(self):
            self.input = "table1.biom"
            self.output = "out_dir"
            self.correl_method = "spearman"
            self.p_adjust = None
            self.outlier_removal = False
            self.verbose = False
            self.force = False
            self.min_sample = None
            self.sparcc_filter = True
            self.procs = 1
            self.min_p = None
            self.min_r = .35

    return Arguments()


# integration test
def test_within_correls_classic_correlation_min_r_min_sample(args1, tmpdir):
    table = simulate_correls()
    loc = tmpdir.mkdir("with_correls_test")
    table.to_json("madebyme", open(str(loc)+"/table1.biom", 'w'))
    os.chdir(str(loc))
    within_correls(args1)
    files = os.listdir(str(loc)+'/out_dir')
    assert "collapsed.biom" in files
    assert "modules.txt" in files
    assert "modules" in files
    assert "conetwork.gml" in files
    assert "correls.txt" in files


# integration test
def test_within_correls_classic_correlation_min_r_sparcc_filter(args2, tmpdir):
    table = simulate_correls()
    loc = tmpdir.mkdir("with_correls_test")
    table.to_json("madebyme", open(str(loc)+"/table1.biom", 'w'))
    os.chdir(str(loc))
    within_correls(args2)
    files = os.listdir(str(loc)+'/out_dir')
    assert "collapsed.biom" in files
    assert "modules.txt" in files
    assert "modules" in files
    assert "conetwork.gml" in files
    assert "correls.txt" in files



from setuptools import setup, find_packages

__author__ = 'shafferm'
__version__ = '0.3.1'

setup(
      name="SCNIC",
      version=__version__,
      setup_requires=['pytest-runner'],
      test_require=['pytest'],
      install_requires=["numpy", "scipy", "networkx", "biom-format", "pandas", "fast_sparCC <= 0.1.6",
                        "scikit-bio == 0.4.2"],
      scripts=["scripts/SCNIC_analysis.py"],
      packages=find_packages(),
      description="A tool for finding and summarizing modules of highly correlated observations in compositional data",
      author="Michael Shaffer",
      author_email='michael.shaffer@ucdenver.edu',
      url="https://github.com/shafferm/SCNIC/",
      download_url="https://github.com/shafferm/SCNIC/tarball/%s" % __version__
)



import numpy as np
from scipy.stats import spearmanr, pearsonr, kendalltau
import warnings
from functools import partial
import pandas as pd


def cor_to_dist(cor):
    # convert from correlation to distance
    return 1 - ((cor + 1) / 2)

def between_correls_from_tables(table1, table2, correl_method=spearmanr, nprocs=1):
    """Take two biom tables and correlation"""
    correls = list()

    if nprocs == 1:
        for data_i, otu_i, _ in table1.iter(axis="observation"):
            for data_j, otu_j, _ in table2.iter(axis="observation"):
                corr = correl_method(data_i, data_j)
                correls.append([otu_i, otu_j, corr[0], corr[1]])
    else:
        import multiprocessing
        if nprocs > multiprocessing.cpu_count():
            warnings.warn("nprocs greater than CPU count, using all avaliable CPUs")
            nprocs = multiprocessing.cpu_count()

        pool = multiprocessing.Pool(nprocs)
        for data_i, otu_i, _ in table1.iter(axis="observation"):
            datas_j = (data_j for data_j, _, _ in table2.iter(axis="observation"))
            corr = partial(correl_method, b=data_i)
            corrs = pool.map(corr, datas_j)
            correls += [(otu_i, table2.ids(axis="observation")[i], corrs[i][0], corrs[i][1])
                        for i in xrange(len(corrs))]
        pool.close()
        pool.join()

    return pd.DataFrame(correls, columns=['feature1', 'feature2', 'r', 'p'])



from __future__ import division

import multiprocessing
import warnings
from collections import Counter
import numpy as np
from biom import Table
from functools import partial
from scipy.spatial.distance import jaccard, braycurtis, euclidean, canberra
import pandas as pd


def cscore(u, v):
    """calculates standardized c-score according to https://en.wikipedia.org/wiki/Checkerboard_score"""
    u = u.astype(bool)
    v = v.astype(bool)
    r_u = np.sum(u)
    r_v = np.sum(v)
    s_uv = np.sum(np.logical_and(u, v))
    return (r_u-s_uv)*(r_v-s_uv)/(r_u+r_v-s_uv)


def paired_distances_from_table(table, dist_metric='braycurtis'):
    """Takes a biom table and finds distances between all pairs of otus"""
    dist_methods = {'jaccard': jaccard, 'cscore': cscore, 'braycurtis': braycurtis, 'euclidean': euclidean,
                    'canberra': canberra}
    dist_metric = dist_methods[dist_metric]
    dists = list()

    for (data_i, otu_i, _), (data_j, otu_j, _) in table.iter_pairwise(axis='observation'):
        dist = dist_metric(data_i, data_j)
        dists.append([str(otu_i), str(otu_j), dist])

    header = ['feature1', 'feature2', 'dist']
    return pd.DataFrame(dists, columns=header)


def refill_biom(row):
    tot = np.sum(row)
    rands = np.random.randint(0, len(row), size=tot)
    counts = Counter(rands)
    new_data = np.zeros(len(row))
    for val in counts:
        new_data[val] = counts[val]
    return new_data


def shuffle_table(table):
    matrix = table.matrix_data.todense()
    np.random.shuffle(matrix)
    return Table(matrix, table.ids(axis="observation"), table.ids())


def shuffle_table_matrix(sparse_matrix):
    matrix = sparse_matrix.todense()
    np.random.shuffle(matrix)
    return Table(matrix, range(matrix.shape[0]), range(matrix.shape[1]))


def refill_table(table):
    matrix = table.matrix_data.todense()
    new_matrix = np.zeros(matrix.shape)
    sums = np.sum(matrix, axis=1)
    for i in xrange(sums.shape[0]):
        rands = np.random.randint(0, matrix.shape[1], size=sums[i])
        counts = Counter(rands)
        for val in counts:
            new_matrix[i, val] = counts[val]
    return Table(new_matrix, table.ids(axis="observation"), table.ids())


def refill_table_matrix(sparse_matrix):
    matrix = sparse_matrix.todense()
    new_matrix = np.zeros(matrix.shape)
    sums = np.sum(matrix, axis=1)
    for i in xrange(sums.shape[0]):
        rands = np.random.randint(0, matrix.shape[1], size=sums[i])
        counts = Counter(rands)
        for val in counts:
            new_matrix[i, val] = counts[val]
    return Table(new_matrix, range(new_matrix.shape[0]), range(new_matrix.shape[1]))


def bootstrapped_distance(bootstrap, measured_dists, sparse_matrix, dist_metric):
    bootstrapped_table = shuffle_table_matrix(sparse_matrix)
    dists = paired_distances_from_table(bootstrapped_table, dist_metric=dist_metric)
    return dists['dist'] < measured_dists


def bootstrap_distance_vals(table, dist_metric='braycurtis', nprocs=1, bootstraps=1000):
    """"""
    dists = paired_distances_from_table(table, dist_metric=dist_metric)
    measured_dists = dists['dist']

    if nprocs == 1:
        print "Using 1 processor to calculate distances"
        multi_results = np.zeros((bootstraps, len(dists)), dtype=bool)
        for i in xrange(bootstraps):
            multi_results[i] = bootstrapped_distance(None, measured_dists, table.matrix_data, dist_metric)

    else:
        if nprocs > multiprocessing.cpu_count():
            warnings.warn("nprocs greater than CPU count, using all avaliable CPUs")
            nprocs = multiprocessing.cpu_count()

        pool = multiprocessing.Pool(nprocs)
        print "Number of processors used: " + str(nprocs)

        pfun = partial(bootstrapped_distance, measured_dists=measured_dists, sparse_matrix=table.matrix_data,
                       dist_metric=dist_metric)
        multi_results = pool.map(pfun, range(bootstraps))
        pool.close()
        pool.join()

    dists['p'] = np.sum(multi_results, axis=0)/bootstraps

    return dists



import pytest
from SCNIC.general import simulate_correls, get_metadata_from_table, filter_table, sparcc_paper_filter,\
                          bonferroni_adjust, bh_adjust, Logger, biom_to_df
from biom.table import Table
import numpy as np
from numpy.testing import assert_allclose
import pandas as pd

# TODO: simulate sparse table to test filtering
# TODO: include HMP table to test filtering?


@pytest.fixture()
def biom_table1():
    table1 = simulate_correls()
    assert isinstance(table1, Table)
    return table1


@pytest.fixture()
def biom_table2():
    arr = np.array([[250,   0, 100, 446,   75],
                    [  0,   0,   1,   1,    2],
                    [  2,   2,   2,   2,    2],
                    [100, 100, 500,   1, 1000],
                    [500,   5,   0,  50,  100]])
    obs_ids = ["otu_%s" % i for i in xrange(5)]
    samp_ids = ["samp_%s" % i for i in xrange(5)]
    return Table(arr, obs_ids, samp_ids)


@pytest.fixture()
def dataframe():
    arr = np.array([[250,   0, 100, 446,   75],
                    [  0,   0,   1,   1,    2],
                    [  2,   2,   2,   2,    2],
                    [100, 100, 500,   1, 1000],
                    [500,   5,   0,  50,  100]])
    arr = arr.transpose()
    return pd.DataFrame(arr, index=["samp_%s" % i for i in xrange(5)], columns=["otu_%s" % i for i in xrange(5)])


@pytest.fixture()
def unadj_ps():
    return [.01, .05, .5]


def test_Logger(tmpdir):
    loc = tmpdir.mkdir("test")
    log_path = str(loc) + "/log.txt"
    logger = Logger(log_path)
    logger["Testing"] = "1, 2, 3"
    logger.output_log()
    log = open(log_path).readlines()
    assert len(log) == 4
    assert log[0].startswith('start time')
    assert log[1].startswith('Testing: 1, 2, 3')
    assert log[-2].startswith('finish time')
    assert log[-1].startswith('elapsed time')


def test_biom_to_df(biom_table2, dataframe):
    df = biom_to_df(biom_table2)
    assert np.array_equal(df.as_matrix(), dataframe.as_matrix())
    assert df.columns.equals(dataframe.columns)
    assert df.index.equals(dataframe.index)


def test_get_metadata_from_table(biom_table1):
    metadata = get_metadata_from_table(biom_table1)
    assert isinstance(metadata, dict)


def test_filter_table(biom_table1):
    table_filt = filter_table(biom_table1, min_samples=10)
    assert isinstance(table_filt, Table)


def test_filter_better(biom_table2):
    table_filt = filter_table(biom_table2, min_samples=4)
    assert len(table_filt.ids(axis="observation")) == 4
    assert len(table_filt.ids(axis="sample")) == 5


def test_sparcc_paper_filter(biom_table1):
    table_filt = sparcc_paper_filter(biom_table1)
    assert isinstance(table_filt, Table)


def test_sparcc_paper_filter_better(biom_table2):
    table_filt = sparcc_paper_filter(biom_table2)
    assert len(table_filt.ids(axis="observation")) == 4
    assert len(table_filt.ids(axis="sample")) == 3


def test_bonferroni_adjust(unadj_ps):
    adj_ps = np.array([.03, .15, 1.5])
    bon_ps = bonferroni_adjust(unadj_ps)
    assert isinstance(bon_ps, np.ndarray)
    assert_allclose(adj_ps, bon_ps)


def test_bh_adjust(unadj_ps):
    adj_ps = np.array([.03, .075, .5])
    bh_ps = bh_adjust(unadj_ps)
    assert isinstance(bh_ps, np.ndarray)
    assert_allclose(adj_ps, bh_ps)



# TODO: Testing correls as list vs as pandas dataframe for speed

from __future__ import division

import os

import networkx as nx
import pandas as pd
import numpy as np

from biom import load_table
from sparcc_fast.sparcc_functions import basis_corr
from sparcc_fast.utils import df_to_correls
from scipy.stats import spearmanr, pearsonr, kendalltau
from scipy.spatial.distance import squareform

import general
import correlation_analysis as ca
import module_maker as mm


def within_correls(args):
    logger = general.Logger("SCNIC_log.txt")
    logger["SCNIC analysis type"] = "within"

    # sanity check args
    if args.min_r is not None and args.min_p is not None:
        raise ValueError("arguments min_p and min_r may not be used concurrently")

    # correlation and p-value adjustment methods
    correl_methods = {'spearman': spearmanr, 'pearson': pearsonr, 'kendall': kendalltau, 'sparcc': None}
    p_methods = {'bh': general.bh_adjust, 'bon': general.bonferroni_adjust}
    correl_method = correl_methods[args.correl_method.lower()]
    if args.p_adjust is not None:
        p_adjust = p_methods[args.p_adjust]
    else:
        p_adjust = None

    # get features to be correlated and extract metadata
    table = load_table(args.input)
    logger["input table"] = args.input
    if args.verbose:
        print "Table loaded: " + str(table.shape[0]) + " observations"
        print ""
    logger["number of samples in input table"] = table.shape[1]
    logger["number of observations in input table"] = table.shape[0]

    # make new output directory and change to it
    if args.output is not None and not os.path.isdir(args.output):
        os.makedirs(args.output)
        os.chdir(args.output)
        logger["output directory"] = args.output

    # filter
    if args.sparcc_filter is True:
        table_filt = general.sparcc_paper_filter(table)
        if args.verbose:
            print "Table filtered: " + str(table_filt.shape[0]) + " observations"
            print ""
        logger["sparcc paper filter"] = True
        logger["number of observations present after filter"] = table_filt.shape[0]
    elif args.min_sample is not None:
        table_filt = general.filter_table(table, args.min_sample)
        if args.verbose:
            print "Table filtered: " + str(table_filt.shape[0]) + " observations"
            print ""
        logger["min samples present"] = args.min_sample
        logger["number of observations present after filter"] = table_filt.shape[0]
    else:
        table_filt = table

    logger["number of processors used"] = args.procs

    # correlate features
    if args.min_r is not None:
        if correl_method in [spearmanr, pearsonr, kendalltau]:
            # calculate correlations
            if args.verbose:
                print "Correlating with " + args.correl_method
            # correlate feature
            cor, p_vals = correl_method(general.biom_to_df(table_filt))
            cor = squareform(cor, checks=False)
        else:
            if args.verbose:
                print "Correlating using sparcc"
            cor, _ = basis_corr(general.biom_to_df(table_filt))
            cor = squareform(cor, checks=False)
        dist = ca.cor_to_dist(cor)
        min_dist = ca.cor_to_dist(args.min_r)
    elif args.min_p is not None:
        raise NotImplementedError()
    else:
        raise ValueError("min_p and min_r not given, one or other needs to be set")
    logger["distance metric used"] = args.correl_method

    if args.verbose:
        print "Features Correlated"
        print ""

    # make modules
    modules = mm.make_modules(dist, min_dist, obs_ids=table_filt.ids(axis="observation"))
    logger["number of modules created"] = len(modules)
    if args.verbose:
        print "Modules Formed"
        print "number of modules: %s" % len(modules)
        print "number of observations in modules: %s" % np.sum([len(i) for i in modules])
        print ""
    mm.write_modules_to_file(modules)

    # collapse modules
    coll_table = mm.collapse_modules(table, modules)
    mm.write_modules_to_dir(table, modules)
    logger["number of observations in output table"] = coll_table.shape[0]
    if args.verbose:
        print "Table Collapsed"
        print "collapsed Table Observations: " + str(coll_table.shape[0])
        print ""
    coll_table.to_json('make_modules.py', open('collapsed.biom', 'w'))

    # print correls and make correlation network
    correls = df_to_correls(pd.DataFrame(squareform(cor), index=table_filt.ids(axis="observation"),
                                         columns=table_filt.ids(axis="observation")))
    if 'p' in correls.columns:
        correls['p-adj'] = p_adjust(correls['p'])
    correls.to_csv('correls.txt', sep='\t', index=False)
    metadata = general.get_metadata_from_table(table_filt)
    net = general.correls_to_net(correls, conet=True, metadata=metadata, min_p=args.min_p, min_r=args.min_r)
    for i, otus in enumerate(modules):
        for otu in otus:
            net.node[otu]["module"] = i
    if args.verbose:
        print "Network Generated"
        print "number of nodes: " + str(net.number_of_nodes())
        print "number of edges: " + str(net.number_of_edges())
    logger["number of nodes"] = net.number_of_nodes()
    logger["number of edges"] = net.number_of_edges()
    nx.write_gml(net, 'conetwork.gml')

    logger.output_log()



import pytest
from SCNIC.general import simulate_correls
from SCNIC.correlation_analysis import between_correls_from_tables
import pandas as pd
from scipy.misc import comb


@pytest.fixture()
def biom_table1():
    return simulate_correls()


# TODO: Induce between table correlations to try to detect
@pytest.fixture()
def biom_table2():
    return simulate_correls()


def test_between_correls_from_tables_single(biom_table1, biom_table2):
    correls = between_correls_from_tables(biom_table1, biom_table2)
    assert isinstance(correls, pd.DataFrame)
    assert correls.shape[0] == biom_table1.shape[0] * biom_table2.shape[0]


def test_between_correls_from_tables_multi(biom_table1, biom_table2):
    correls = between_correls_from_tables(biom_table1, biom_table2, nprocs=2)
    assert isinstance(correls, pd.DataFrame)
    assert correls.shape[0] == biom_table1.shape[0] * biom_table2.shape[0]
